from composer_xls import ComposerXLS

import os
import tempfile
import logging

log = logging.getLogger(__name__)

class DocumentWriter(object):
    "runs a composer"
   

    composer = None

    def writestr(self):
        if not self.composer:
            # default format is XLS
            self.composer = ComposerXLS(self.data_model, self.grid, self)
            log.info("Setting output format to XLS")
    

        return self.composer.run()


    def write(self, filename):
        ext = os.path.splitext(filename)[-1].lower()
        if ext == "xls":
            self.composer = ComposerXLS(self.data_model, self.grid, self)
            log.info("Setting output format to XLS, based on file extension")


        f = open(filename, "w")
        f.write(self.writestr())
        f.close()
        log.info("wrote file: %s" % filename)

    def write_gdata(self, name, username, password, auth_token=None):
        try:
            import gdata
            import gdata.docs.service
        except ImportError:
            raise Exception("Must install package 'gdata' to use write_gdata()")

        tmp_file, tmp_file_path = tempfile.mkstemp()
        self.write(tmp_file_path)

        gd_client = gdata.docs.service.DocsService()
        gd_client.ssl = True
        if not auth_token:
            gd_client.ClientLogin(
                         username,
                         password,
                         "marmir-1.0")
        else:
            #TODO: use the token
            raise Exception("oauth not yet supported")

        ms =  gdata.MediaSource(file_path=tmp_file_path, 
                                        content_type='application/vnd.ms-excel')
        entry = gd_client.Upload(ms, name)

        #cleanup
        os.unlink(tmp_file_path)

        return gd_client.GetClientLoginToken()




import logging
log = logging.getLogger(__name__)


class DjangoGrid(object):

    def populate(self, indata):
        for required in ('row_count', 'col_count', 'headers'):
            if not hasattr(self, required):
                raise Exception("missing required attribute to Grid: %s" % required)
        # create a grid
        self.grid_data = [[None] * self.col_count for i in range(self.row_count)]

        # now populate
        # this is pass one
        # want to do as much processing here as we can
        # we populate left to right, top to bottom
        for row_id in range(self.row_count):
            row = indata[row_id]
            for col_id in range(self.col_count):
                field_type_class = self.headers[col_id]
                data = getattr(row, self.titles[col_id])
                self.grid_data[row_id][col_id] = field_type_class(data)

                #Ilport pdb; pdb.set_trace()
        log.info("populated grid %sX%s" % (self.row_count, self.col_count))



from grid_base import GridBase


class Serializer(object):
    """ Class that pairs data with models """
    def __init__(self, data_model, data, grid_class=GridBase):
        self.data = data
        self.data_model = data_model
        self.grid_class = grid_class

    def serialize(self):
        """ returnes serialzed data into a Grid """
        grid = self.grid_class()

        # we care about order
        # that was set in model_base
        grid.row_count = len(self.data)
        field_headers =  self.data_model.field_headers
        grid.col_count = len(field_headers)
        grid.headers = field_headers 
        grid.titles = self.data_model.field_titles
        grid.populate(self.data)

        return grid




from lib.font_data.decorators import memoized

excel_color_dict = {}

excel_color_dict['0, 0, 0'] = 0x08
excel_color_dict['255, 255, 255'] = 0x09
excel_color_dict['255, 0, 0'] = 0x0A
excel_color_dict['0, 255, 0'] = 0x0B
excel_color_dict['0, 0, 255'] = 0x0C
excel_color_dict['255, 255, 0'] = 0x0D
excel_color_dict['255, 0, 255'] = 0x0E
excel_color_dict['0, 255, 255'] = 0x0F
excel_color_dict['128, 0, 0'] = 0x10
excel_color_dict['0, 128, 0'] = 0x11
excel_color_dict['0, 0, 128'] = 0x12
excel_color_dict['128, 128, 0'] = 0x13
excel_color_dict['128, 0, 128'] = 0x14
excel_color_dict['0, 128, 128'] = 0x15
excel_color_dict['192, 192, 192'] = 0x16
excel_color_dict['128, 128, 128'] = 0x17
excel_color_dict['153, 153, 255'] = 0x18
excel_color_dict['153, 51, 102'] = 0x1A
excel_color_dict['255, 255, 204'] = 0x1C
excel_color_dict['204, 255, 255'] = 0x1D
excel_color_dict['102, 0, 102'] = 0x1E
excel_color_dict['255, 128, 128'] = 0x1F
excel_color_dict['0, 102, 204'] = 0x28
excel_color_dict['204, 204, 255'] = 0x29
excel_color_dict['0, 204, 255'] = 0x2A
excel_color_dict['204, 255, 204'] = 0x2B
excel_color_dict['255, 255, 153'] = 0x2C
excel_color_dict['153, 204, 255'] = 0x2D
excel_color_dict['255, 153, 204'] = 0x2E
excel_color_dict['204, 153, 255'] = 0x2F
excel_color_dict['255, 204, 153'] = 0x30
excel_color_dict['51, 102, 255'] = 0x31
excel_color_dict['51, 204, 204'] = 0x32
excel_color_dict['153, 204, 0'] = 0x33
excel_color_dict['255, 204, 0'] = 0x34
excel_color_dict['255, 153, 0'] = 0x35
excel_color_dict['255, 102, 0'] = 0x36
excel_color_dict['102, 102, 153'] = 0x37
excel_color_dict['150, 150, 150'] = 0x38
excel_color_dict['0, 51, 102'] = 0x39
excel_color_dict['51, 153, 102'] = 0x3A
excel_color_dict['0, 51, 0'] = 0x3B
excel_color_dict['51, 51, 0'] = 0x3C
excel_color_dict['153, 51, 0'] = 0x3D
excel_color_dict['51, 51, 153'] = 0x3E
excel_color_dict['51, 51, 51'] = 0x3F
  

@memoized
def rgb(c):
    split = (c[0:2], c[2:4], c[4:6])
    out = []
    for x in split:
        out.append(int(x,16)) 
    return out
  
@memoized  
def get_closest_rgb_match(hex):
    hex = hex.replace("#",'').strip()
    color_dict = excel_color_dict
    orig_rgb = rgb(hex)
    new_color = ''
    min_distance = 195075
    orig_r = orig_rgb[0]
    orig_g = orig_rgb[1]
    orig_b = orig_rgb[2]
    for key in color_dict.iterkeys():
        new_r = int(key.split(',')[0])
        new_g = int(key.split(',')[1])
        new_b = int(key.split(',')[2])
        r_distance = orig_r - new_r
        g_distance = orig_g - new_g
        b_distance = orig_b - new_b
        current_distance = (r_distance * r_distance) + (g_distance * g_distance) + (b_distance * b_distance)
        if current_distance < min_distance:
            min_distance = current_distance
            new_color = key
    return color_dict.get(new_color)
      




from composer_base import ComposerBase
import lib.xlwt_0_7_2 as xlwt
from lib.font_data.core import get_string_width
from lib.xldate.convert import to_excel_from_C_codes
import logging
import StringIO
import model_base
import style_base
import color_converter

log = logging.getLogger(__name__)


def get_string_width_from_style(char_string, style):
    point_size = style.font.height / 0x14  # convert back to points
    font_name = style.font.name
    if not font_name:
        font_name = 'Arial'
    return int(get_string_width(font_name, point_size, char_string) * 50)


class styleXLS(style_base.StyleBase):

    font_points = 12

    def get_pattern(self):
        pattern = xlwt.Pattern()
        pattern.pattern = 1
        if self.background_color:
            color = color_converter.get_closest_rgb_match(self.background_color)
        else:
            color = 1

        pattern.pattern_fore_colour = color
        return pattern

    def get_font_color(self):
        color = 0
        if self.color:
            color = color_converter.get_closest_rgb_match(self.color)
        return color

    def get_border(self):
        border = xlwt.Borders()
        if False:  # TODO borders
            border.left = border.right = border.top = border.bottom = 3000
            if self.border_color:
                color = color_converter.get_closest_rgb_match(self.border_color)
                border.top_color = color
                border.bottom_color = color
                border.left_color = color
                border.right_color = color
        return border

    def is_bold(self):
        if self.font_style == 'bold':
            return True
        return False

    def get_font_points(self):
        if self.font_size:
            return self.font_size
        return 12  # TODO: default from config?

    def get_font_name(self):
        if not self.font_family:
            return 'Arial'
        return self.font_family

    def get_text_align(self):
        text_align = xlwt.Alignment()
        # HORZ - (0-General, 1-Left, 2-Center, 3-Right, 4-Filled, 5-Justified, 6-CenterAcrossSel, 7-Distributed)
        horz = 0
        if self.text_align == 'center':
            horz = 2
        elif self.text_align == 'right':
            horz = 3
        elif self.text_align == 'left':
            horz = 1  # left
        else:
            log.warn("Unknown text_align %s" % self.text_align)

        text_align.horz = horz
        return text_align


class ComposerXLS(ComposerBase):

    def convert_style(self, stylestr):
        in_style = styleXLS()
        in_style.style_from_string(stylestr)

        style = xlwt.XFStyle()
        fnt1 = xlwt.Font()
        fnt1.name = in_style.get_font_name()
        fnt1.bold = in_style.is_bold()
        fnt1.height = in_style.get_font_points() * 0x14
        fnt1.colour_index = in_style.get_font_color()
        style.font = fnt1
        style.alignment = in_style.get_text_align()
        style.pattern = in_style.get_pattern()
        style.borders = in_style.get_border()

        return style

    def cell_to_value(self, cell, row_id):

        if self.document.config.headers and row_id == 0:
            css_like_style = self.document.config.header_style
        elif len(self.document.config.row_styles) == 0:
            css_like_style = ''
        else:
            style_index = row_id % len(self.document.config.row_styles)
            css_like_style = self.document.config.row_styles[style_index]

        style = self.convert_style(css_like_style)

        if type(cell) == model_base.HeaderFieldType:
            style = self.convert_style(self.document.config.header_style)
            return cell.data, style

        elif type(cell) in (model_base.IntFieldType, model_base.StringFieldType):
            return cell.data, style

        elif type(cell) == model_base.DateTimeFieldType:
            style.num_format_str = self.document.config.get('datetime_format', 'M/D/YY h:mm')
            return cell.data, style
        elif type(cell) == model_base.DateFieldType:
            num_string_format = self.document.config.get('date_format', 'M/D/YY')
            if cell.format:
                num_string_format = to_excel_from_C_codes(cell.format, self.document.config)
            style.num_format_str = num_string_format
            return cell.data, style

        else:
            return cell.data, style

    def start_new_row(self, id):
        pass

    def end_row(self, id):
        pass

    def write_cell(self, row_id, col_id, cell):

        value, style = self.cell_to_value(cell, row_id)
        if type(cell) == model_base.ImageFieldType:
            if cell.width:
                self.sheet.col(col_id).width = cell.width * 256
            if cell.height:
                self.sheet.col(col_id).height = cell.height * 256
            self.sheet.insert_bitmap(value, row_id, col_id)

        elif type(cell) == model_base.URLFieldType:
            self.sheet.write(
                row_id,
                col_id,
                xlwt.Formula('HYPERLINK("%s";"%s")' % (value, cell.displayname)),
                style
            )

        else:
            # most cases
            self.sheet.write(row_id, col_id, value, style)
        self.done_write_cell(row_id, col_id, cell, value, style)

    def done_write_cell(self, row_id, col_id, cell, value, style):

        if self.document.config.get('adjust_all_col_width', False):

            current_width = self.sheet.col_width(col_id) + 0x0d00
            log.info("current width is %s" % current_width)
            new_width = None

            if type(cell) == model_base.StringFieldType:
                new_width = get_string_width_from_style(value, style)

            elif type(cell) == model_base.DateTimeFieldType:
                new_width = 6550  # todo: different date formats

            elif type(cell) == model_base.URLFieldType:
                new_width = get_string_width_from_style(cell.displayname, style)

            if new_width and new_width > current_width:
                log.info("setting col #%s form width %s to %s" % (col_id, current_width, new_width))
                col = self.sheet.col(col_id)
                if new_width > 65535:  # USHRT_MAX
                    new_width = 65534
                    current_width = new_width
                col.width = new_width

    def set_option(self, key):

        val = getattr(self.document.config, key)
        if key == 'freeze_col' and val and val > 0:
            self.sheet.panes_frozen = True
            self.sheet.vert_split_pos = val

        elif key == 'freeze_row' and val and val > 0:
            self.sheet.panes_frozen = True
            self.sheet.horz_split_pos = val

        else:

            log.info("Nothing to be done for %s" % key)

            return
        log.info("Set option %s" % key)

    def run(self):

        self.w = xlwt.Workbook(style_compression=2)
        self.sheet = self.w.add_sheet('Sheet 1')
        if self.document.config.headers:
            self.write_header()
        self.iterate_grid()
        self.finish()

        # write the file to string
        output = StringIO.StringIO()
        self.w.save(output)
        contents = output.getvalue()
        output.close()

        return contents





class ConfigBase(object):
    """ Holds the configuration """

    def get(self,key,default=None):
        if hasattr(self,key):
            return getattr(self,key)
        return default
    
    def __init__(self, config_dict=None):
        if config_dict:
            self.set_dict(config_dict)

    def set_dict(self, config_dict):
        for k,v in config_dict.items():
            setattr(self, k, v)


    # default settings
    headers = True
    header_style = "color: #ffffff; font-family: arial; background-color: #0000B3; font-size: 12pt; text-align: center"
    freeze_col = 0
    freeze_row = 1
    row_styles = (
        "color: #000000; font-family: arial; background-color: #666666; border-color: #ff0000",
        "color: #000000; font-family: arial; background-color: #FFFFFF"   # Alternate
    )
    adjust_all_col_width = True
    datetime_format = 'M/D/YY h:mm:ss'
    date_format = 'M/D/YY'
    time_format = "h:mm:ss"






import logging
from mm import model_base
import django.db.models.fields as fields

log = logging.getLogger(__name__)


class DjangoDataModel(object):
    """ Data Model creates a list of system defined data types in self.field_headers"""

    def __init__(self, data, order=None):
        """ constructor takes data as a tuple or list"""

        self.field_headers = []
        self.field_titles = []
        if data.count() == 0:
            raise Exception("Can not make spreadsheets with an empty set")
        first_data = data[0]

        header_types = {}
        for f in first_data._meta.fields:  # TODO: + obj._meta.many_to_many
            header_types[f.name] = self.figure_out_type(f)

        #TODO: 'if order' sort the list

        # assign data
        for verbose_name, field_type_class in header_types.items():

            # we add it to the 'class' so to be
            # used in every instance
            self.field_titles.append(verbose_name)
            self.field_headers.append(field_type_class)
            log.info("created field type %s for %s" % (field_type_class, verbose_name))

    def __serialize(obj):
        obj._meta.fields

    def _string_types(self):
        return [
            fields.CharField,
            fields.CommaSeparatedIntegerField,
            fields.IPAddressField,
            fields.SlugField,
            fields.TextField,
        ]

    def _int_types(self):
        return [
            fields.AutoField,
            fields.IntegerField,
            fields.PositiveIntegerField,
            fields.PositiveSmallIntegerField,
            fields.SmallIntegerField,

        ]

    def _bool_types(self):
        return [
            fields.BooleanField,
            fields.NullBooleanField,
        ]

    def _date_types(self):
        return [
            fields.DateField,
        ]

    def _time_types(self):
        return [
            fields.TimeField
        ]

    def _datetime_types(self):
        return [
            fields.DateTimeField,
        ]

    def _decimal_types(self):
        return [
            fields.DecimalField,
        ]

    def _float_types(self):
        return [
            fields.FloatField,
        ]

    def _none_types(self):
        return [
            fields.NullBooleanField
        ]

    def _url_types(self):
        return [
            fields.URLField
        ]

    def type_mapping(self):
        return (
            (self._string_types, model_base.StringFieldType),
            (self._int_types, model_base.IntFieldType),
            (self._bool_types, model_base.BoolFieldType),
            (self._date_types, model_base.DateFieldType),
            (self._time_types, model_base.TimeFieldType),
            (self._datetime_types, model_base.DateTimeFieldType),
            (self._decimal_types, model_base.DecimalFieldType),
            (self._float_types, model_base.FloatFieldType),
            (self._none_types, model_base.NoneFieldType),
            (self._url_types, model_base.URLFieldType),
        )

    def figure_out_type(self, item):
        """

This is how django stores types in sqlite3:

"AutoField" integer NOT NULL PRIMARY KEY,
"BooleanField" bool NOT NULL,
"CharField" varchar(50) NOT NULL,
"CommaSeparatedIntegerField" varchar(25) NOT NULL,
"DateField" date NOT NULL,
"DateTimeField" datetime NOT NULL,
"DecimalField" decimal NOT NULL,
"EmailField" varchar(75) NOT NULL,
"FloatField" real NOT NULL,
"IntegerField" integer NOT NULL,
"IPAddressField" char(15) NOT NULL,
"NullBooleanField" bool,
"PositiveIntegerField" integer unsigned NOT NULL,
"PositiveSmallIntegerField" smallint unsigned NOT NULL,
"SlugField" varchar(30) NOT NULL,
"SmallIntegerField" smallint NOT NULL,
"TextField" text NOT NULL,
"TimeField" time NOT NULL,
"URLField" varchar(100) NOT NULL
        """
        item_type = type(item)
        for func, mm_type in self.type_mapping():
            if item_type in func():
                return mm_type

        log.warn("Returning None type for type %s" % item_type)
        return model_base.NoneFieldType



import unittest
import convert


class TestsConvert(unittest.TestCase):

    def test_formats(self):
        tests = (
            ("%b %d %H:%M:%S %Y", "mmm dd hh:mm:ss yyyy"), # Jul 08 08:08:10 2011
            ('%b|%B|%c|%d|%f|%H|%m|%M|%p|%S|%x|%X|%y|%Y|%%','mmm|mmmm|M/D/YY h:mm:ss|dd|[ss].00|hh|mm|mm|AM/PM|ss|M/D/YY|h:mm:ss|yy|yyyy|\%'),
        )
        for test in tests:
            excel = convert.to_excel_from_C_codes(test[0],{})
            self.assertEqual(test[1], excel)


        tests = (
            ('%a', '%A', '%I', '%j', '%U', '%w', '%W', '%z', '%Z',)
        )
        for test in tests:
            self.assertRaises(convert.UnsupportedFormatCodeException, 
                                 convert.to_excel_from_C_codes, test, {})

if __name__ == "__main__":
    unittest.main()






import logging
from model_base import HeaderFieldType

log = logging.getLogger(__name__)

class ComposerBase(object):
    """ Used by Composers """
    def run(self):
       raise Exception("Overwrite run() in subclass") 

    def __init__(self,data_model, grid, document):
        self.data_model = data_model
        self.grid = grid
        self.document = document
        self.row_id = 0
        self.col_id = 0


    def iterate_grid(self):

        for row in self.grid.grid_data:
            self.col_id = 0
            self.start_new_row(self.row_id)
            for cell in row:
                self.write_cell(self.row_id, self.col_id, cell)
                self.col_id += 1
            self.end_row(self.row_id)
            self.row_id += 1

        

    def write_header(self):
        i = 0
        for header in self.data_model.field_titles:
            cell = HeaderFieldType(data=header)
            log.info(cell.__dict__)
            self.write_cell(0, i, cell)
            i += 1
        self.row_id += 1



    def finish(self):
        """ Things we do after we are done """
        for key in [x for x in dir(self.document.config) if not x.startswith("_") ]:
            self.set_option(key)







# -*- coding: windows-1252 -*-

import sys
import struct

class Reader:
    def __init__(self, filename, dump = False):
        self.dump = dump
        self.STREAMS = {}

        doc = file(filename, 'rb').read()
        self.header, self.data = doc[0:512], doc[512:]
        del doc

        self.__build_header()
        self.__build_MSAT()
        self.__build_SAT()
        self.__build_directory()
        self.__build_short_sectors_data()
        
        if len(self.short_sectors_data) > 0:
            self.__build_SSAT()
        else:
            if self.dump and (self.total_ssat_sectors != 0 or self.ssat_start_sid != -2):
                print 'NOTE: header says that must be', self.total_ssat_sectors, 'short sectors'
                print 'NOTE: starting at', self.ssat_start_sid, 'sector'
                print 'NOTE: but file does not contains data in short sectors'
            self.ssat_start_sid = -2
            self.total_ssat_sectors = 0
            self.SSAT = [-2]

        for dentry in self.dir_entry_list[1:]:
            (did, 
             sz, name, 
             t, c, 
             did_left, did_right, did_root, 
             dentry_start_sid, 
             stream_size
            ) = dentry
            stream_data = ''
            if stream_size > 0:
                if stream_size >= self.min_stream_size:
                    args = (self.data, self.SAT, dentry_start_sid, self.sect_size)
                else:
                    args = (self.short_sectors_data, self.SSAT, dentry_start_sid, self.short_sect_size)
                stream_data = self.get_stream_data(*args)

            if name != '':
                # BAD IDEA: names may be equal. NEED use full paths...
                self.STREAMS[name] = stream_data

    
    def __build_header(self):
        self.doc_magic             = self.header[0:8]

        if self.doc_magic != '\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1':
            raise Exception, 'Not an OLE file.'

        self.file_uid              = self.header[8:24]
        self.rev_num               = self.header[24:26]
        self.ver_num               = self.header[26:28]
        self.byte_order            = self.header[28:30]
        self.log2_sect_size,       = struct.unpack('<H', self.header[30:32])
        self.log2_short_sect_size, = struct.unpack('<H', self.header[32:34])
        self.total_sat_sectors,    = struct.unpack('<L', self.header[44:48])
        self.dir_start_sid,        = struct.unpack('<l', self.header[48:52])
        self.min_stream_size,      = struct.unpack('<L', self.header[56:60])
        self.ssat_start_sid,       = struct.unpack('<l', self.header[60:64])
        self.total_ssat_sectors,   = struct.unpack('<L', self.header[64:68])
        self.msat_start_sid,       = struct.unpack('<l', self.header[68:72])
        self.total_msat_sectors,   = struct.unpack('<L', self.header[72:76])
         
        self.sect_size        = 1 << self.log2_sect_size
        self.short_sect_size  = 1 << self.log2_short_sect_size

        if self.dump:
            print 'file magic: '
            print_bin_data(self.doc_magic)

            print 'file uid: '
            print_bin_data(self.file_uid)

            print 'revision number: '
            print_bin_data(self.rev_num)
         
            print 'version number: '
            print_bin_data(self.ver_num)
            
            print 'byte order: '
            print_bin_data(self.byte_order)
            
            print 'sector size                                :', hex(self.sect_size), self.sect_size
            #print 'total sectors in file                      :', hex(self.total_sectors), self.total_sectors
            print 'short sector size                          :', hex(self.short_sect_size), self.short_sect_size
            print 'Total number of sectors used for the SAT   :', hex(self.total_sat_sectors), self.total_sat_sectors
            print 'SID of first sector of the directory stream:', hex(self.dir_start_sid), self.dir_start_sid
            print 'Minimum size of a standard stream          :', hex(self.min_stream_size), self.min_stream_size
            print 'SID of first sector of the SSAT            :', hex(self.ssat_start_sid), self.ssat_start_sid
            print 'Total number of sectors used for the SSAT  :', hex(self.total_ssat_sectors), self.total_ssat_sectors
            print 'SID of first additional sector of the MSAT :', hex(self.msat_start_sid), self.msat_start_sid
            print 'Total number of sectors used for the MSAT  :', hex(self.total_msat_sectors), self.total_msat_sectors


    def __build_MSAT(self):
        self.MSAT = list(struct.unpack('<109l', self.header[76:]))
        
        next = self.msat_start_sid
        while next > 0:
           msat_sector = struct.unpack('<128l', self.data[next*self.sect_size:(next+1)*self.sect_size])
           self.MSAT.extend(msat_sector[:127])
           next = msat_sector[-1]

        if self.dump:
            print 'MSAT (header part): \n', self.MSAT[:109]
            print 'additional MSAT sectors: \n', self.MSAT[109:]


    def __build_SAT(self):
        sat_stream = ''.join([self.data[i*self.sect_size:(i+1)*self.sect_size] for i in self.MSAT if i >= 0])

        sat_sids_count = len(sat_stream) >> 2
        self.SAT = struct.unpack('<%dl' % sat_sids_count, sat_stream) # SIDs tuple

        if self.dump:
            print 'SAT sid count:\n', sat_sids_count
            print 'SAT content:\n', self.SAT


    def __build_SSAT(self):
        ssat_stream = self.get_stream_data(self.data, self.SAT, self.ssat_start_sid, self.sect_size)

        ssids_count = len(ssat_stream) >> 2
        self.SSAT = struct.unpack('<%dl' % ssids_count, ssat_stream)

        if self.dump:
            print 'SSID count:', ssids_count
            print 'SSAT content:\n', self.SSAT


    def __build_directory(self):
        dir_stream = self.get_stream_data(self.data, self.SAT, self.dir_start_sid, self.sect_size)

        self.dir_entry_list = []

        i = 0
        while i < len(dir_stream):
            dentry = dir_stream[i:i+128] # 128 -- dir entry size
            i += 128
            
            did = len(self.dir_entry_list)
            sz, = struct.unpack('<H', dentry[64:66])
            if sz > 0 :
                name = dentry[0:sz-2].decode('utf_16_le', 'replace')
            else:
                name = u''
            t,  = struct.unpack('B', dentry[66])
            c,  = struct.unpack('B', dentry[67])
            did_left ,  = struct.unpack('<l', dentry[68:72])
            did_right ,  = struct.unpack('<l', dentry[72:76])
            did_root ,  = struct.unpack('<l', dentry[76:80])
            dentry_start_sid ,  = struct.unpack('<l', dentry[116:120])
            stream_size ,  = struct.unpack('<L', dentry[120:124])

            self.dir_entry_list.extend([(did, sz, name, t, c, 
                                            did_left, did_right, did_root, 
                                            dentry_start_sid, stream_size)]) 

        if self.dump:
            dentry_types = {
                0x00: 'Empty',
                0x01: 'User storage',
                0x02: 'User stream',
                0x03: 'LockBytes',
                0x04: 'Property',
                0x05: 'Root storage'
            }
            node_colours = {
                0x00: 'Red',
                0x01: 'Black'
            }
            print 'total directory entries:', len(self.dir_entry_list)

            for dentry in self.dir_entry_list:
                (did, sz, name, t, c, 
                 did_left, did_right, did_root, 
                 dentry_start_sid, stream_size) = dentry
                print 'DID', did
                print 'Size of the used area of the character buffer of the name:', sz
                print 'dir entry name:', repr(name)
                print 'type of entry:', t, dentry_types[t]
                print 'entry colour:', c, node_colours[c]
                print 'left child DID :', did_left
                print 'right child DID:', did_right
                print 'root DID       :', did_root
                print 'start SID       :', dentry_start_sid
                print 'stream size     :', stream_size
                if stream_size == 0:
                    print 'stream is empty'
                elif stream_size >= self.min_stream_size:
                    print 'stream stored as normal stream'
                else:
                    print 'stream stored as short-stream'

    
    def __build_short_sectors_data(self):
        (did, sz, name, t, c, 
         did_left, did_right, did_root, 
         dentry_start_sid, stream_size) = self.dir_entry_list[0]
        assert t == 0x05 # Short-Stream Container Stream (SSCS) resides in Root Storage
        if stream_size == 0:
            self.short_sectors_data = ''
        else:
            self.short_sectors_data = self.get_stream_data(self.data, self.SAT, dentry_start_sid, self.sect_size)


    def get_stream_data(self, data, SAT, start_sid, sect_size):
        sid = start_sid
        chunks = [(sid, sid)]
        stream_data = ''

        while SAT[sid] >= 0:
            next_in_chain = SAT[sid]
            last_chunk_start, last_chunk_finish = chunks[-1]
            if next_in_chain == last_chunk_finish + 1:
                chunks[-1] = last_chunk_start, next_in_chain
            else:
                chunks.extend([(next_in_chain, next_in_chain)]) 
            sid = next_in_chain
        for s, f in chunks:
            stream_data += data[s*sect_size:(f+1)*sect_size]
        #print chunks
        return stream_data

        
def print_bin_data(data):
    i = 0
    while i < len(data):
        j = 0
        while (i < len(data)) and (j < 16):
            c = '0x%02X' % ord(data[i])
            sys.stdout.write(c)
            sys.stdout.write(' ')
            i += 1
            j += 1
        print
    if i == 0:
        print '<NO DATA>'



# This implementation writes only 'Root Entry', 'Workbook' streams
# and 2 empty streams for aligning directory stream on sector boundary
# 
# LAYOUT:
# 0         header
# 76                MSAT (1st part: 109 SID)
# 512       workbook stream
# ...       additional MSAT sectors if streams' size > about 7 Mb == (109*512 * 128)
# ...       SAT
# ...       directory stream
#
# NOTE: this layout is "ad hoc". It can be more general. RTFM

class XlsDoc:
    SECTOR_SIZE = 0x0200
    MIN_LIMIT   = 0x1000

    SID_FREE_SECTOR  = -1
    SID_END_OF_CHAIN = -2
    SID_USED_BY_SAT  = -3
    SID_USED_BY_MSAT = -4

    def __init__(self):
        #self.book_stream = ''                # padded
        self.book_stream_sect = []

        self.dir_stream = ''
        self.dir_stream_sect = []

        self.packed_SAT = ''
        self.SAT_sect = []

        self.packed_MSAT_1st = ''
        self.packed_MSAT_2nd = ''
        self.MSAT_sect_2nd = []

        self.header = ''

    def __build_directory(self): # align on sector boundary
        self.dir_stream = ''

        dentry_name      = '\x00'.join('Root Entry\x00') + '\x00'
        dentry_name_sz   = len(dentry_name)
        dentry_name_pad  = '\x00'*(64 - dentry_name_sz)
        dentry_type      = 0x05 # root storage
        dentry_colour    = 0x01 # black
        dentry_did_left  = -1
        dentry_did_right = -1
        dentry_did_root  = 1
        dentry_start_sid = -2
        dentry_stream_sz = 0

        self.dir_stream += struct.pack('<64s H 2B 3l 9L l L L',
           dentry_name + dentry_name_pad,
           dentry_name_sz,
           dentry_type,
           dentry_colour,
           dentry_did_left, 
           dentry_did_right,
           dentry_did_root,
           0, 0, 0, 0, 0, 0, 0, 0, 0,
           dentry_start_sid,
           dentry_stream_sz,
           0
        )

        dentry_name      = '\x00'.join('Workbook\x00') + '\x00'
        dentry_name_sz   = len(dentry_name)
        dentry_name_pad  = '\x00'*(64 - dentry_name_sz)
        dentry_type      = 0x02 # user stream
        dentry_colour    = 0x01 # black
        dentry_did_left  = -1
        dentry_did_right = -1
        dentry_did_root  = -1
        dentry_start_sid = 0     
        dentry_stream_sz = self.book_stream_len

        self.dir_stream += struct.pack('<64s H 2B 3l 9L l L L',
           dentry_name + dentry_name_pad,
           dentry_name_sz,
           dentry_type,
           dentry_colour,
           dentry_did_left, 
           dentry_did_right,
           dentry_did_root,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 
           dentry_start_sid,
           dentry_stream_sz,
           0
        )
        
        # padding
        dentry_name      = ''
        dentry_name_sz   = len(dentry_name)
        dentry_name_pad  = '\x00'*(64 - dentry_name_sz)
        dentry_type      = 0x00 # empty
        dentry_colour    = 0x01 # black
        dentry_did_left  = -1
        dentry_did_right = -1
        dentry_did_root  = -1
        dentry_start_sid = -2
        dentry_stream_sz = 0

        self.dir_stream += struct.pack('<64s H 2B 3l 9L l L L',
           dentry_name + dentry_name_pad,
           dentry_name_sz,
           dentry_type,
           dentry_colour,
           dentry_did_left, 
           dentry_did_right,
           dentry_did_root,
           0, 0, 0, 0, 0, 0, 0, 0, 0,
           dentry_start_sid,
           dentry_stream_sz,
           0
        ) * 2
    
    def __build_sat(self):
        # Build SAT
        book_sect_count = self.book_stream_len >> 9
        dir_sect_count  = len(self.dir_stream) >> 9
        
        total_sect_count     = book_sect_count + dir_sect_count
        SAT_sect_count       = 0
        MSAT_sect_count      = 0
        SAT_sect_count_limit = 109
        while total_sect_count > 128*SAT_sect_count or SAT_sect_count > SAT_sect_count_limit:
            SAT_sect_count   += 1
            total_sect_count += 1
            if SAT_sect_count > SAT_sect_count_limit:
                MSAT_sect_count      += 1
                total_sect_count     += 1
                SAT_sect_count_limit += 127


        SAT = [self.SID_FREE_SECTOR]*128*SAT_sect_count

        sect = 0
        while sect < book_sect_count - 1:
            self.book_stream_sect.append(sect)
            SAT[sect] = sect + 1
            sect += 1
        self.book_stream_sect.append(sect)
        SAT[sect] = self.SID_END_OF_CHAIN
        sect += 1

        while sect < book_sect_count + MSAT_sect_count:
            self.MSAT_sect_2nd.append(sect)
            SAT[sect] = self.SID_USED_BY_MSAT
            sect += 1

        while sect < book_sect_count + MSAT_sect_count + SAT_sect_count:
            self.SAT_sect.append(sect)            
            SAT[sect] = self.SID_USED_BY_SAT
            sect += 1

        while sect < book_sect_count + MSAT_sect_count + SAT_sect_count + dir_sect_count - 1:
            self.dir_stream_sect.append(sect)
            SAT[sect] = sect + 1
            sect += 1
        self.dir_stream_sect.append(sect)
        SAT[sect] = self.SID_END_OF_CHAIN
        sect += 1

        self.packed_SAT = struct.pack('<%dl' % (SAT_sect_count*128), *SAT)

        MSAT_1st = [self.SID_FREE_SECTOR]*109
        for i, SAT_sect_num in zip(range(0, 109), self.SAT_sect):
            MSAT_1st[i] = SAT_sect_num
        self.packed_MSAT_1st = struct.pack('<109l', *MSAT_1st)

        MSAT_2nd = [self.SID_FREE_SECTOR]*128*MSAT_sect_count
        if MSAT_sect_count > 0:
            MSAT_2nd[- 1] = self.SID_END_OF_CHAIN

        i = 109
        msat_sect = 0
        sid_num = 0
        while i < SAT_sect_count:
            if (sid_num + 1) % 128 == 0:
                #print 'link: ',
                msat_sect += 1
                if msat_sect < len(self.MSAT_sect_2nd):
                    MSAT_2nd[sid_num] = self.MSAT_sect_2nd[msat_sect]
            else:
                #print 'sid: ',
                MSAT_2nd[sid_num] = self.SAT_sect[i]
                i += 1
            #print sid_num, MSAT_2nd[sid_num]
            sid_num += 1

        self.packed_MSAT_2nd = struct.pack('<%dl' % (MSAT_sect_count*128), *MSAT_2nd)

        #print vars()
        #print zip(range(0, sect), SAT)
        #print self.book_stream_sect
        #print self.MSAT_sect_2nd
        #print MSAT_2nd
        #print self.SAT_sect
        #print self.dir_stream_sect


    def __build_header(self):
        doc_magic             = '\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1'
        file_uid              = '\x00'*16
        rev_num               = '\x3E\x00'
        ver_num               = '\x03\x00'
        byte_order            = '\xFE\xFF'
        log_sect_size         = struct.pack('<H', 9)
        log_short_sect_size   = struct.pack('<H', 6)
        not_used0             = '\x00'*10
        total_sat_sectors     = struct.pack('<L', len(self.SAT_sect))
        dir_start_sid         = struct.pack('<l', self.dir_stream_sect[0])
        not_used1             = '\x00'*4        
        min_stream_size       = struct.pack('<L', 0x1000)
        ssat_start_sid        = struct.pack('<l', -2)
        total_ssat_sectors    = struct.pack('<L', 0)

        if len(self.MSAT_sect_2nd) == 0:
            msat_start_sid        = struct.pack('<l', -2)
        else:
            msat_start_sid        = struct.pack('<l', self.MSAT_sect_2nd[0])

        total_msat_sectors    = struct.pack('<L', len(self.MSAT_sect_2nd))

        self.header =       ''.join([  doc_magic,
                                        file_uid,
                                        rev_num,
                                        ver_num,
                                        byte_order,
                                        log_sect_size,
                                        log_short_sect_size,
                                        not_used0,
                                        total_sat_sectors,
                                        dir_start_sid,
                                        not_used1,
                                        min_stream_size,
                                        ssat_start_sid,
                                        total_ssat_sectors,
                                        msat_start_sid,
                                        total_msat_sectors
                                    ])
                                        

    def save(self, file_name_or_filelike_obj, stream):
        # 1. Align stream on 0x1000 boundary (and therefore on sector boundary)
        padding = '\x00' * (0x1000 - (len(stream) % 0x1000))
        self.book_stream_len = len(stream) + len(padding)

        self.__build_directory()
        self.__build_sat()
        self.__build_header()
        
        f = file_name_or_filelike_obj
        we_own_it = not hasattr(f, 'write')
        if we_own_it:
            f = open(file_name_or_filelike_obj, 'wb')
        f.write(self.header)
        f.write(self.packed_MSAT_1st)
        f.write(stream)
        f.write(padding)
        f.write(self.packed_MSAT_2nd)
        f.write(self.packed_SAT)
        f.write(self.dir_stream)
        if we_own_it:
            f.close()



import re


class UnsupportedFormatCodeException(Exception):
    pass

def to_excel_from_C_codes(cdate_str, config):
    """
    ref http://office.microsoft.com/en-us/excel-help/create-a-custom-number-format-HP010342372.aspx
    and http://docs.python.org/2/library/datetime.html

    """
    pairs = (

    ('%a', None, "abbreviated weekday names require special function in excel"), # not supported
	('%A', None, "weeday naes require a special funtion in excel"), # "
	('%b', 'mmm'), # month as an abbreviation (Jan to Dec).
	('%B', 'mmmm'), # month as a full name (January to December)
	('%c', config.get('datetime_format', 'M/D/YY h:mm:ss') ), # date and time representation.
	('%d', 'dd'),
	('%f', '[ss].00'), #  Microsecond as a decimal number [0,999999], zero-padded on the left || Elapsed time (seconds and hundredths)  3735.80 [ss].00
	('%H', 'hh'), # Hours   00-23   hh
	('%I', None, "AM or PM required for 12 hour clock"), # AM or PM required
	('%j', None, "Day of year not supported in Excel"),
	('%m', 'mm'),
	('%M', 'mm'),
	('%p', 'AM/PM'), # 
	('%S', 'ss'),
	('%U', None, "Excel has no support for week number"),
	('%w', None, "Excel has no support for week day"),
	('%W', None, "Excel has no support for Week Number of year"),
	('%x', config.get('datetime_format', 'M/D/YY') ), # date 
	('%X', config.get('time_format', 'h:mm:ss') ), # time
	('%y', 'yy'), #  year as a two-digit number.
	('%Y', 'yyyy'), #  year as a four-digit number.
	('%z', None, "Excel has no time zone support"),
	('%Z', None, "Excel has no time zone support"),
	('%%', r'\%'),
)

    original_str = cdate_str
    for t in pairs:    
        if not t[1] and cdate_str.find(t[0]) > -1:
            reason = "Excel does not support"
            if len(t) > 2:
                reason = t[2]
            raise UnsupportedFormatCodeException("Could not replace %s (%s) found in %s" % (t[0], 
                                                                      reason, 
                                                               original_str))
        elif not t[1]:
            continue

        cdate_str = re.sub(t[0], t[1], cdate_str)

    return cdate_str










import collections
import functools

class memoized(object):
   '''Decorator. Caches a function's return value each time it is called.
   If called later with the same arguments, the cached value is returned
   (not reevaluated).
   '''
   def __init__(self, func):
      self.func = func
      self.cache = {}
   def __call__(self, *args):
      if not isinstance(args, collections.Hashable):
         # uncacheable. a list, for instance.
         # better to not cache than blow up.
         return self.func(*args)
      if args in self.cache:
         return self.cache[args]
      else:
         value = self.func(*args)
         self.cache[args] = value
         return value
   def __repr__(self):
      '''Return the function's docstring.'''
      return self.func.__doc__
   def __get__(self, obj, objtype):
      '''Support instance methods.'''
      return functools.partial(self.__call__, obj)



import logging
from datetime import datetime
import inspect
import sys
from lib.font_data.decorators import memoized

log = logging.getLogger(__name__)


class BaseFieldType(object):

    def __init__(self, data):
        self.data = data

    def __repr__(self):
        return unicode(self)

    def __unicode__(self):
        return u"%s" % type(self)


class HeaderFieldType(BaseFieldType):
    pass


class DateFieldType(BaseFieldType):
    def __init__(self, data, format=None):
        self.format = format
        super(DateFieldType, self).__init__(data)


class TimeFieldType(BaseFieldType):
    pass


class DateTimeFieldType(BaseFieldType):

    def __init__(self, data):
        if data and data.tzinfo:
            data = data.replace(tzinfo=None)  # excel can't handle
        super(DateTimeFieldType, self).__init__(data)


class IntFieldType(BaseFieldType):
    pass


class FloatFieldType(BaseFieldType):
    pass


class DecimalFieldType(BaseFieldType):
    pass


class StringFieldType(BaseFieldType):
    pass


class BoolFieldType(BaseFieldType):
    pass


class URLFieldType(BaseFieldType):
    def __init__(self, path, displayname=None):
        if not displayname:
            displayname = path
        self.displayname = displayname
        super(URLFieldType, self).__init__(path)


class ImageFieldType(BaseFieldType):
    def __init__(self, path, width=None, height=None):
        self.width = width
        self.height = height
        super(ImageFieldType, self).__init__(path)


class NoneFieldType(BaseFieldType):
    pass


@memoized
def get_members_list():
    return [
        x[1] for x
        in inspect.getmembers(sys.modules[__name__], inspect.isclass)
        if issubclass(x[1], BaseFieldType)]


def is_custom_mm_type(inst):
    members = get_members_list()
    if type(inst) in members:
        return True
    return False


class DataModel(object):
    """ Data Model creates a list of system defined data types in self.field_headers"""

    def __init__(self, data, order=None):
        """ constructor takes data as a tuple or list"""
        self.field_headers = []
        self.field_titles = []
        if len(data) == 0:
            raise Exception("Can not make spreadsheets with an empty set")
        first_data = data[0]
        if type(first_data) == dict and len(first_data) > 0:
            if order:
                # add in this order it was explicitly set
                self.field_titles = order
            else:
                # no order set, just get
                self.field_titles = first_data.keys()

            for k in self.field_titles:
                log.info("looking at %s ..." % data[0][k])
                if is_custom_mm_type(data[0][k]):
                    field_type_class = type(data[0][k])
                else:
                    # we figure out the type
                    field_type_class = self.figure_out_type(data[0][k])

                # we add it to the 'class' so to be
                # used in every instance
                self.field_headers.append(field_type_class)
                log.info("created field type %s for %s" % (field_type_class, k))

    def figure_out_type(self, item):
        item_type = type(item)
        if item_type == unicode or item_type == str:
            return StringFieldType

        elif item_type == int:
            return IntFieldType

        elif item_type == datetime:
            return DateTimeFieldType

        log.warn("Returning None type for type %s" % item_type)
        return NoneFieldType



import unittest
import core


class test_font_data(unittest.TestCase):

    def test_checkwidth1(self):

        width = core.get_string_width('Arial', 11, 'hello world') 
        self.assertEqual(width, 52.421875)

    def test_checkwidth2(self):
        width = core.get_string_width('Times New Roman', 23 ,'The quick brown fox jumps over the lazy dog')
        self.assertEqual(width, 420.46875)







#!/usr/bin/env python


__version__ = "0.10"
__author__ = [
    "Brian Ray <brianhray@gmail.com>",
]
__license__ = "TBD"

from document_base import *
import config_base
import model_base

Date = model_base.DateFieldType
URL = model_base.URLFieldType
Image = model_base.ImageFieldType

Config = config_base.ConfigBase




import logging as log

class StyleBase(object):

    color = None
    font_family = None
    background_color = None
    border_color = None
    font_style = None
    text_align = None
    _font_size_points = None # points

    @property
    def font_size(self):
        return self._font_size_points

    @font_size.setter
    def font_size(self, value):
        if 'pt' in value:
            value = int(value.replace('pt','')) 
        elif 'px' in value:
            log.warning("font-size does not (yet) support pixel sizes") # TODO: support pixel
            self._font_size_points = None
            return
        else:
            log.warning("assuming font-size is in points")

        self._font_size_points = int(value)

    def style_from_string(self, in_str):
        
        for attr in in_str.split(";"):
            if attr.strip() == '': continue
            k,v = attr.split(":")
            value = v.strip()
            key = k.strip().replace("-", "_")
            if not hasattr(self, key):
                log.warn("Unknown style attribute %s: %s" % (key,value))
                continue

            setattr(self, key, value)




from setuptools import setup, find_packages


setup(
    name='Marmir',
    version="0.1.1",
    description='Marmir: makes spreadsheets easy',
    author='Brian Ray',
    author_email='brianhray@gmail.com',
    maintainer='Brian Ray',
    maintainer_email='brianhray@gmail.com',
    url='https://github.com/brianray/mm',
    long_description="Python power spreadsheets on steroids",
    #      license="Brian Ray", TBD
    platforms=["any"],
    packages=find_packages(),
    include_package_data=True,
)



from document_writers import DocumentWriter
from model_base import DataModel
from config_base import ConfigBase
from serializer_base import Serializer
from grid_base import GridBase
import logging

log = logging.getLogger(__name__)

class Document(DocumentWriter):
    """
    Document reporesents the abstact view you interact with in order to send
    data for your document and ultimately get output.
    """


    def __init__(self, 
                    data, 
                    data_model_class=None, 
                    grid_class=None,
                    serializer_class=None, 
                    config=None, 
                    config_dict=None,
                    order=None):
        """      
         data -- a dict or a list of data you wish to use for a the
                 spreadsheet
         data_model -- (optional) fields defenitions
         grid_model -- (optional) takes data_model and data and fills grid
         serializer_class -- (optional) class to use to serialize raw data
         config -- (optional) Configuration (ConfigBase) instance
         config_dict -- (optional) a dictionary of key/values of settings
        """
        self.data = data
        self.config = config
        if not self.config:
            self.config = ConfigBase()
        if config_dict:
            self.config.set_dict(config_dict)
 
        # make a data model if one does not exist
        self.data_model_class = data_model_class
        if not data_model_class:
            self.data_model_class = DataModel

        self.data_model = self.data_model_class(data, order=order)

        # grid base 
        if not grid_class:
            grid_class = GridBase

        # Serialize the data       
        # we look at it here once and only once
        # we look at it again when we write
        # goal to pass over data no more than twice, if possible
        if not serializer_class:
            serializer_class = Serializer
        serializer = serializer_class(self.data_model, 
                                        self.data,
                                        grid_class=grid_class
                                        )

        # returns a grid instance
        self.grid = serializer.serialize()
       
        log.info("Documnet Created")
 
    def set_composer(self, composer):
        self.composer = composer
        



import os.path as path
from decorators import memoized
try:
    import cPickle as pickle
except:
    import pickle

LATEST_FONT_DATA = "font_data_ms_fonts.bin"

@memoized
def get_font_data():
    this_file_path = path.abspath(__file__)
    data_path = path.join( path.dirname(this_file_path), LATEST_FONT_DATA)
    pkl_file = open(data_path, 'rb')
    FONT_DATA = pickle.load(pkl_file)
    pkl_file.close()
    return FONT_DATA


@memoized
def get_character_data(font_name, char):
    font_data = get_font_data()
    font_name = font_name.replace("-","_").replace(" ","_")
    if font_name not in font_data:
        font_name = font_name.capitalize()
        if font_name not in font_data:
            raise Exception("no font data for font %s" % font_name)
    font_set = font_data[font_name]
    if char not in font_set['values']:
        return font_set['default_width'], {} # empty kerns
    return font_set['values'][char]

@memoized
def get_character_width(font_name, char):
    width, kerns = get_character_data(font_name, char)
    return width

@memoized
def get_kern_offset(font_name, char1, char2):
    width, kerns = get_character_data(font_name, char1)
    if char2 in kerns:
        return kerns[char2]
    return 0

@memoized
def get_string_width(font_name, point_size, char_string):
    out_width_256 = 0
    current_pos = 0
    str_length = len(char_string)
    for char in char_string:
        out_width_256 += get_character_width(font_name, char)
        if current_pos != (str_length-1):
            out_width_256 += get_kern_offset(font_name, char, char_string[current_pos+1])
        current_pos += 1
    return out_width_256 * ( point_size / 256.0 )

if __name__ == "__main__":
    get_character_data('Arial', 'A')
